/*
 * Address Conversion Tool
 *
 * Organization : OGCIO(E)11A
 * Last modified: 10/06/2004
 *
 * @author TH HEUNG
 * @version 1.0
*/

package hksarg.ogcio.AddressConversion;

import java.lang.*;
import java.text.BreakIterator;
import java.util.*;
import java.util.regex.PatternSyntaxException;
import hksarg.ogcio.AddressConversion.Tokenizer.*;

public class FreeTextAddress
{
	protected Vector addressLines;
	protected int maxLineWidth;
	protected boolean isAbbreviated;
	protected boolean isAllLineTokenized;
	protected boolean forceBreakBetweenWord;
	
	public FreeTextAddress()
	{
		addressLines = new Vector();
		maxLineWidth = 0;
		isAbbreviated = false;
		isAllLineTokenized = false;
		forceBreakBetweenWord = false;
	}
	
	public FreeTextAddress(String[] source)
	{
		addressLines = new Vector();
		maxLineWidth = 0;
		isAbbreviated = false;
		isAllLineTokenized = false;
		readAddress(source);
		forceBreakBetweenWord = false;
	}
	
	public FreeTextAddress(FreeTextAddress source)
	{
		String[] sourceLine = source.getAddress();
		addressLines = new Vector(sourceLine.length);
		for (int i=0;i<sourceLine.length;i++)
			addressLines.add(new String(sourceLine[i]));
		maxLineWidth = source.getMaxWidth();
		isAbbreviated = source.isAbbreviate();
		isAllLineTokenized = source.isTokenize();
		forceBreakBetweenWord = source.isForceBreakBetweenWord();
	}
	
	private void readAddress(String[] source)
	{
		for (int i=0;i<source.length;i++)
		{
			String line = source[i].trim();
			int pos = line.indexOf("\n");
			while (pos>0)
			{
				String firstPortion = line.substring(0,pos).trim();
				String secondPortion = line.substring(pos).trim();
				if (firstPortion.length()>0)
				{
					addressLines.add(removeContinuousSpace(firstPortion));
					if (firstPortion.length() > maxLineWidth)
						maxLineWidth = firstPortion.length();
				}
				line = secondPortion;
				pos = line.indexOf("\n");
			}
			addressLines.add(removeContinuousSpace(line));
			if (line.length() > maxLineWidth)
				maxLineWidth = line.length();
		}
	}
	
	protected String removeContinuousSpace(String source)
	{
		String resultStr = source;
		try
		{
			resultStr = resultStr.replaceAll("\\s{2,}"," ");
		}
		catch (PatternSyntaxException e)
		{
		}
		return resultStr;
	}
	
	protected void countMaxLineWidth()
	{
		maxLineWidth = 0;
		for (int i=0;i<addressLines.size();i++)
		{
			String line = (String)addressLines.elementAt(i);
			if (maxLineWidth < line.length())
				maxLineWidth = line.length();
		}
	}
	
	public void mergeLines(int lineNum1, int lineNum2)
	throws ArrayIndexOutOfBoundsException
	{
		String line1 = mergeAddressString(
				(String)addressLines.elementAt(lineNum1),
				(String)addressLines.elementAt(lineNum2));
		addressLines.setElementAt(line1, lineNum1);
		addressLines.removeElementAt(lineNum2);
		if (line1.length() > maxLineWidth)
			maxLineWidth = line1.length();
	}
	
	public void replaceLine(String newLine, int lineNum)
	throws ArrayIndexOutOfBoundsException
	{
		String line = newLine.trim();
		if (line.length() == 0)
			addressLines.removeElementAt(lineNum);
		else if (lineNum >= addressLines.size())
			addressLines.add(line);
		else
			addressLines.setElementAt(line, lineNum);
		countMaxLineWidth();
	}
	
	public void removeLine(int lineNum)
	throws ArrayIndexOutOfBoundsException
	{
		addressLines.removeElementAt(lineNum);
		countMaxLineWidth();
	}
	
	public void abbreviate()
	{
		if (isAbbreviated)
			return;
		Abbreviator abbreviator = Abbreviator.getInstance();
		for (int i=0;i<addressLines.size();i++)
		{
			String line = (String)addressLines.elementAt(i);
			line = abbreviator.abbreviate(line);
			addressLines.setElementAt(line,i);
		}
		isAbbreviated = true;
		countMaxLineWidth();
	}
	
	public void tokenize()
	{
		if (isAllLineTokenized)
			return;
		
		Vector array = new Vector();
		for (int i=0;i<addressLines.size();i++)
		{
			String[] token = getToken(i);
			for (int j=0;j<token.length;j++)
				array.add(token[j].trim());
		}
		if (array.size() > 0)
			addressLines = array;
		isAllLineTokenized = true;
	}
	
	public String[] getAddress()
	{
		return (String[])addressLines.toArray(new String[addressLines.size()]);
	}
	
	public int getLineTotal()
	{
		return addressLines.size();
	}
	
	public int getMaxWidth()
	{
		return maxLineWidth;
	}
	
	public String getLine(int lineNum)
	throws ArrayIndexOutOfBoundsException
	{
		return (String)addressLines.elementAt(lineNum);
	}
	
	public boolean isAbbreviate()
	{
		return isAbbreviated;
	}
	
	public boolean isTokenize()
	{
		return isAllLineTokenized;
	}
	
	public boolean isForceBreakBetweenWord()
	{
		return forceBreakBetweenWord;
	}
	
	public void setForceBreakBetweenWord()
	{
		forceBreakBetweenWord = true;
		isAllLineTokenized = false;
	}
	
	public String[] getToken(int lineNum)
	throws ArrayIndexOutOfBoundsException
	{
		return getToken(lineNum,false);
	}
	
	public String[] getToken(int lineNum, boolean breakBetweenWord)
	throws ArrayIndexOutOfBoundsException
	{
		String str = getLine(lineNum);
		String[] tokens;
		if (forceBreakBetweenWord || breakBetweenWord)
		{
			tokens = breakWord(str);
		}
		else
		{
			FuzzyTokenizer tokenizer = new FuzzyTokenizer();
			tokens = tokenizer.tokenizer(str);
			if (FreeTextAddressConvertor.DEBUG)
			{
				printToken(tokens);
			}
		}
		
		return tokens;
	}
	
	private String[] breakWord(String line)
	throws ArrayIndexOutOfBoundsException
	{
		BreakIterator boundary = BreakIterator.getLineInstance();
		boundary.setText(line);
		int start = boundary.first();
		Vector array = new Vector();
		for (int end = boundary.next();
				end != BreakIterator.DONE;
				start = end, end = boundary.next())
		{
			array.add(line.substring(start,end));
		}
		return (String[])array.toArray(new String[array.size()]);
	}
	
	private void printToken(String[] tokens)
	{
		for (int i=0;i<tokens.length;i++)
		{
			System.out.println("Token: \"" + tokens[i] + "\"");
		}
	}

	public static boolean isChineseUnicodeBlock(char c)
	{
		Character.UnicodeBlock block = Character.UnicodeBlock.of(c);
		if (block == Character.UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS)
			return true;
		if (block == Character.UnicodeBlock.CJK_UNIFIED_IDEOGRAPHS_EXTENSION_A)
			return true;
		return false;
	}
	
	public static String mergeAddressString(String lineA, String lineB)
	{
		if (lineB==null)
			return lineA.trim();
		if (lineB.length()==0)
			return lineA.trim();
		
		if (lineA.matches(".*?(\\s|-)"))
			return lineA.concat(lineB).trim();
		
		if (FreeTextAddress.isChineseUnicodeBlock(lineA.charAt(lineA.length()-1))
				&& FreeTextAddress.isChineseUnicodeBlock(lineB.charAt(0)))
			return lineA.concat(lineB).trim();
		
		return lineA.concat(" ").concat(lineB).trim();
	}
}
